@@ -60,36 +60,55 @@ class EthCall(object):
             data = utils.hex2bytes(data)
         selector = data[:4]
 
+        sig = None
         if Web3.isAddress(to_or_sig):
             to = to_or_sig
             sigs = Signatures(self.scan.get_abi(to))
             sig = sigs.find_by_selector(selector)
-            assert sig, "No method match 0x%s" % selector.hex()
-        else:
+
+        elif type(to_or_sig) is str:
             sig = Signature.from_sig(to_or_sig)
-            assert sig.inputs, "Invalid sig: %s" % to_or_sig
         
-        print("Method:")
-        print(' ', sig)
-        
-        args = sig.decode_args(data)
-        if args:
-            print("Arguments:")
+        if sig is None or sig.selector != selector: 
+            # Invalid sig from ABI or user provided sig.
+            sig_str = utils.selector_to_sigs(selector, True)
+            if sig_str:
+                sig = Signature.from_sig(sig_str)
+
+        if sig is None or sig.selector != selector: # Still invalid.
+            print("No sig found for selector 0x%s." % selector.hex())
+            data = data[4:]
+            if not data: # No data.
+                return
+            print("Guessing types ...")
             i = 0
-            for name, typ in sig.inputs:
-                if name is None:
-                    name = 'arg%d' % (i+1)
-                value = args[i]
-
-                if isinstance(value, bytes):
-                    value = value.hex()
-                elif Web3.isAddress(value):
-                    value = self.scan.get_address_name(value)
-
-                print(' ', "%s %s = %s" %(typ, name, value))
+            for offset, typ, value in utils.guess_calldata_types(data):
+                print("[%d] +%s   %s   %s" % (i, offset, typ, value))
                 i += 1
+            return
+
         else:
-            print("No args.")
+            print("Method:")
+            print(' ', sig)
+            
+            args = sig.decode_args(data)
+            if args:
+                print("Arguments:")
+                i = 0
+                for name, typ in sig.inputs:
+                    if name is None:
+                        name = 'arg%d' % (i+1)
+                    value = args[i]
+
+                    if isinstance(value, bytes):
+                        value = value.hex()
+                    elif Web3.isAddress(value):
+                        value = self.scan.get_address_name(value)
+
+                    print(' ', "%s %s = %s" %(typ, name, value))
+                    i += 1
+            else:
+                print("No args.")
 
     def eth_call(self, to, sig_or_name, args=[], sender=None, throw_on_revert=False, **kwargs):
         """
