@@ -3,6 +3,7 @@ import os
 import atexit
 import re
 from typing import Any, Dict
+from pytest import skip
 
 import requests
 from web3 import Web3
@@ -81,16 +82,32 @@ class SelectorDatabase(object):
             logger.debug("Load sig db from %s" % SIG_DB_PATH)
 
         atexit.register(self.save)
-        
-    def get_sig_from_selector(self, selector, only_one=False, online=True):
+    
+    def _normalize_selector(self, selector, with_prefix=False):
+        """
+        Return hex format.
+        """
         if type(selector) is int:
             selector = '%08x' % selector
+        elif type(selector) is bytes:
+            selector = selector.hex()
         else:
             selector = str(selector).lower()
-            if selector.startswith('0x'):
-                selector = selector[2:]
         
-        # No 0x prefix
+        if selector.startswith('0x'):
+            if not with_prefix:
+                selector = selector[2:]
+        else:
+            if with_prefix:
+                selector = '0x' + selector
+
+        return selector
+
+    def get_sig_from_selector(self, selector, only_one=False, online=True):
+
+        selector = self._normalize_selector(selector, False)
+        
+        # No 0x prefix in DB.
         if selector in self.db:
             sigs = self.db[selector]
             if type(sigs) is str:
@@ -125,10 +142,7 @@ class SelectorDatabase(object):
 
     def get_sig_online(self, selector, only_one=False):
         try:
-            if type(selector) is int:
-                selector = '%08x' % selector
-            else:
-                selector = str(selector).lower()
+            selector = self._normalize_selector(selector, True)
 
             # Get sig from sam's db.
 
@@ -187,6 +201,9 @@ class SelectorDatabase(object):
 
 
 def selector_to_sigs(selector, only_one=False):
+    """
+    selector: with 0x prefix.
+    """
     db = SelectorDatabase.get()
     return db.get_sig_from_selector(selector, only_one)
 
@@ -239,38 +256,114 @@ def guess_calldata_types(data):
     data: hex data without selector.
     return [(type, value),..]
     """
-    buf = hex2bytes(data)
+    if type(data) is str:
+        buf = hex2bytes(data)
+    else:
+        buf = data
+
     if len(buf) % 32 != 0:
         buf += b"\x00" * (32 - (len(buf) % 32))
     
     results = []
-    end = len(buf)
     for i in range(len(buf)//32):
-        if i * 32 >= end:
-            break
 
         value = buf[i*32: (i+1)*32]
         uint256 = int.from_bytes(value, 'big')
 
-        if uint256 < len(buf) - 32:
+        if uint256 < len(buf) - 32 and uint256 > i * 32:
             offset = uint256
             length_bytes = buf[offset: offset + 32]
             length = int.from_bytes(length_bytes, 'big')
             if offset + 32 + length <= len(buf):
                 bytes_data = buf[offset + 32: offset + 32 + length]
                 if '\\' not in repr(bytes_data): # Printable.
-                    results.append(("string", bytes_data.decode('utf-8')))
+                    results.append((
+                        "%#x" % (i*32), 
+                        "string",
+                        bytes_data.decode('utf-8') + '\t' + "// offset %#x length %d" % (offset, length)
+                    ))
                 else:
-                    results.append(("bytes", str(bytes_data)))
+                    results.append((
+                        "%#x" % (i*32), 
+                        "bytes",  
+                        str(bytes_data) + '\t' + "// offset %#x length %d" % (offset, length)
+                    ))
 
-                end = min(offset, end) # Skip data
-                continue # This is an offset for bytes/string, just continue.
+                continue # This is an offset for bytes/string, just end processing. .
 
         if uint256 < 2**112: # Small value as uint
-            results.append(("uint256", "%d(%#x)" %(uint256, uint256)))
+            results.append(("%#x" % (i*32), "uint256", "%d (%#x)" %(uint256, uint256)))
         elif len('%x' % uint256) in range(29, 41): # 12-18 Prefix zero as address.
-            results.append(("address", '%0#42x' % uint256))
+            results.append(("%#x" % (i*32), "address", '%0#42x' % uint256))
         else:
-            results.append(("unknown", value.hex()))
-    
+            results.append(("%#x" % (i*32), "unknown", value.hex()))
+
     return results
+
+
+class CoinPrice(object):
+
+    single_instance = None
+
+    def __init__(self) -> None:
+
+        self.chain_coin_id = {
+            "eth": "ethereum",
+            "bsc": "binancecoin",
+            "avax": "avalanche-2",
+            "ftm": "fantom",
+            "arbi": "ethereum",
+            "op": "ethereum",
+            "matic": "matic-network"
+        }
+        # arbitrum, avalanche, bsc, celo, ethereum, fantom, optimism, polygon, tron
+        self.chain_id = {
+            "eth": "ethereum",
+            "bsc": "bsc",
+            "arbi": "arbitrum",
+            "avax": "avalanche",
+            "ftm": "fantom",
+            "op": "optimism",
+            "matic": "polygon"
+        }
+
+        self.url = "https://coins.llama.fi/prices/current/"
+
+
+    # https://coins.llama.fi/prices/current/coingecko:binancecoin
+    # {"coins":{"coingecko:binancecoin":{"price":344.74,"symbol":"BNB","timestamp":1667551373,"confidence":0.99}}}
+    
+    def get_native(self, chain):
+        gecko_id = "coingecko:%s" % self.chain_coin_id.get(chain, chain)
+        ret = requests.get(self.url + gecko_id).json()["coins"]
+        if gecko_id in ret:
+            return ret[gecko_id]
+        else:
+            return None
+
+    # https://coins.llama.fi/prices/current/ethereum:0xc02aaa39b223fe8d0a0e5c4f27ead9083c756cc2,ethereum:0x419d0d8bdd9af5e606ae2232ed285aff190e711b
+    # {"coins":{"ethereum:0x419d0d8bdd9af5e606ae2232ed285aff190e711b":{"decimals":8,"symbol":"FUN","price":0.00756368,"timestamp":1667555283,"confidence":0.99},"ethereum:0xc02aaa39b223fe8d0a0e5c4f27ead9083c756cc2":{"decimals":18,"symbol":"WETH","price":1582.85,"timestamp":1667555277,"confidence":0.99}}}
+    
+    def get_token(self, chain, *addresses):
+        chain = self.chain_id.get(chain, chain)
+        tokens = []
+        for address in addresses:
+            tokens.append('%s:%s' % (chain, address))
+        url = self.url + ','.join(tokens)
+        ret = requests.get(url).json()["coins"]
+
+        prices = []
+        for token in tokens:
+            if token in ret:
+                info = ret[token]
+                info["address"] = token.split(':')[1]
+                prices.append(info)
+            else:
+                prices.append(None)
+        return prices
+
+    @classmethod
+    def get(cls):
+        if CoinPrice.single_instance is None:
+            CoinPrice.single_instance = CoinPrice()
+        return CoinPrice.single_instance
\ No newline at end of file
